
/************************************************************************

    NixieDisplay.C - a Nixie tube display widget
    Copyright (C) 2000 Greg Ercolano, erco@3dsite.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include <stdio.h>			// sscanf()
#include <stdlib.h>			// exit()
#include <string.h>			// memset()
#include "NixieDisplay.H"

// THE DIGIT XPM FILES
#include "xpms/invalid.xpm"
#include "xpms/space.xpm"
#include "xpms/slash.xpm"
#include "xpms/dots.xpm"
#include "xpms/dash.xpm"
#include "xpms/d0.xpm"
#include "xpms/d1.xpm"
#include "xpms/d2.xpm"
#include "xpms/d3.xpm"
#include "xpms/d4.xpm"
#include "xpms/d5.xpm"
#include "xpms/d6.xpm"
#include "xpms/d7.xpm"
#include "xpms/d8.xpm"
#include "xpms/d9.xpm"
#include "xpms/a_a.xpm"
#include "xpms/a_b.xpm"
#include "xpms/a_c.xpm"
#include "xpms/a_d.xpm"
#include "xpms/a_e.xpm"
#include "xpms/a_f.xpm"
#include "xpms/a_g.xpm"
#include "xpms/a_h.xpm"
#include "xpms/a_i.xpm"
#include "xpms/a_j.xpm"
#include "xpms/a_k.xpm"
#include "xpms/a_l.xpm"
#include "xpms/a_m.xpm"
#include "xpms/a_n.xpm"
#include "xpms/a_o.xpm"
#include "xpms/a_p.xpm"
#include "xpms/a_q.xpm"
#include "xpms/a_r.xpm"
#include "xpms/a_s.xpm"
#include "xpms/a_t.xpm"
#include "xpms/a_u.xpm"
#include "xpms/a_v.xpm"
#include "xpms/a_w.xpm"
#include "xpms/a_x.xpm"
#include "xpms/a_y.xpm"
#include "xpms/a_z.xpm"

void NixieDisplay::_Init()
{
    box(FL_FLAT_BOX); 
    color(FL_BLACK); 
    callback(_WinQuit_CB, (void*)this);

    // INITIALIZE
    _tdigits = 0;
    _digits[_tdigits++] = new Digit('\x7f', invalid_xpm);
    _digits[_tdigits++] = new Digit('a', a_a_xpm);
    _digits[_tdigits++] = new Digit('b', a_b_xpm);
    _digits[_tdigits++] = new Digit('c', a_c_xpm);
    _digits[_tdigits++] = new Digit('d', a_d_xpm);
    _digits[_tdigits++] = new Digit('e', a_e_xpm);
    _digits[_tdigits++] = new Digit('f', a_f_xpm);
    _digits[_tdigits++] = new Digit('g', a_g_xpm);
    _digits[_tdigits++] = new Digit('h', a_h_xpm);
    _digits[_tdigits++] = new Digit('i', a_i_xpm);
    _digits[_tdigits++] = new Digit('j', a_j_xpm);
    _digits[_tdigits++] = new Digit('k', a_k_xpm);
    _digits[_tdigits++] = new Digit('l', a_l_xpm);
    _digits[_tdigits++] = new Digit('m', a_m_xpm);
    _digits[_tdigits++] = new Digit('n', a_n_xpm);
    _digits[_tdigits++] = new Digit('o', a_o_xpm);
    _digits[_tdigits++] = new Digit('p', a_p_xpm);
    _digits[_tdigits++] = new Digit('q', a_q_xpm);
    _digits[_tdigits++] = new Digit('r', a_r_xpm);
    _digits[_tdigits++] = new Digit('s', a_s_xpm);
    _digits[_tdigits++] = new Digit('t', a_t_xpm);
    _digits[_tdigits++] = new Digit('u', a_u_xpm);
    _digits[_tdigits++] = new Digit('v', a_v_xpm);
    _digits[_tdigits++] = new Digit('w', a_w_xpm);
    _digits[_tdigits++] = new Digit('x', a_x_xpm);
    _digits[_tdigits++] = new Digit('y', a_y_xpm);
    _digits[_tdigits++] = new Digit('z', a_z_xpm);
    _digits[_tdigits++] = new Digit('0', d0);
    _digits[_tdigits++] = new Digit('1', d1);
    _digits[_tdigits++] = new Digit('2', d2);
    _digits[_tdigits++] = new Digit('3', d3);
    _digits[_tdigits++] = new Digit('4', d4);
    _digits[_tdigits++] = new Digit('5', d5);
    _digits[_tdigits++] = new Digit('6', d6);
    _digits[_tdigits++] = new Digit('7', d7);
    _digits[_tdigits++] = new Digit('8', d8);
    _digits[_tdigits++] = new Digit('9', d9);
    _digits[_tdigits++] = new Digit('-', dash);
    _digits[_tdigits++] = new Digit('/', slash);
    _digits[_tdigits++] = new Digit(':', dots);
    _digits[_tdigits++] = new Digit(' ', space);

    speed = 0.0;
    memset(_dig, 0, sizeof(_dig));
    memset(_lastdig, 0, sizeof(_lastdig));
    
    SetString("-- -- -- --");
    UpdateSpeed(1.0);
}

void NixieDisplay::_Tick_CB(void *data)
{
    NixieDisplay *o = (NixieDisplay*)data;
    o->_Tick_CB2();
}

void NixieDisplay::_Tick_CB2()
{
    int diff = 0;
    for ( int t=0; _dig[t] && _lastdig[t] && t<(int)sizeof(_dig); t++ )
        if ( _dig[t] != _lastdig[t] )
	    { diff = 1; }

    /// I YAM HERE: IMPLEMENT FADE ALGORITHM HERE.
    ///             Make an IMAGE BUFFER, such that the old pixels are
    ///             slowly 'washed away' by dividing them away.
    ///             New pixels (or 'on' pixels) should get applied
    ///             similarly
    ///     pixel = pixel * .5;		// fade away old pixel value
    ///     pixel = pixel + newpixel;	// add in new value
    ///     if ( pixel > 1.0 ) pixel = 1;	// don't go > 1
    ///

    if ( diff )
    {
	Fl::repeat_timeout(0.5, _Tick_CB, (void*)this);	// changed? 
    }
    else
    {
	Fl::repeat_timeout(speed, _Tick_CB, (void*)this);
    }
    redraw();
}

// HANDLE EVENTS IN THE WINDOW
int NixieDisplay::handle(int e)
{
    static int xoff = 0, yoff = 0;
    int ret = Fl_Double_Window::handle(e);
    switch ( e )
    {
        // DOWNCLICK IN WINDOW CREATES CURSOR OFFSETS
	case FL_PUSH:
	    xoff = x() - Fl::event_x_root();
	    yoff = y() - Fl::event_y_root();
	    ret = 1;

	case FL_DRAG:
	    // DRAG THE WINDOW AROUND THE SCREEN
	    position(xoff + Fl::event_x_root(), yoff + Fl::event_y_root());
	    redraw();
	    ret = 1;

	case FL_RELEASE:
	    show();		// raise
	    ret =1;

    }
    return(ret);
}

// NUMERIC VALUE (0-9) RETURNS APPROPRIATE RGB DIGIT
Digit *NixieDisplay::Value2Digit(char val)
{
    for ( int t=0; t<_tdigits; t++ )
        if ( _digits[t]->IsChar(val) )
	    return(_digits[t]);
    return(_digits[0]);				// unknown character
}

// DRAW TWO DIGITS COMPED OVER ONE ANOTHER AT X/Y POSITION
//    Returns 1 if comp occurred.
//
int NixieDisplay::_DrawDigitComp(Digit *a, Digit *b, int x, int y)
{
    // DONT COMP UNLESS DIGITS ARE DIFFERENT
    if ( a == b )
    {
        // SAME? JUST DRAW IT
	a->Draw(x, y);
	return(0);
    }
    else
    {
        // DIFFERENT? COMP
	static Digit tmp;
	tmp.Comp(*a, *b, 1.0, 1.0);
	tmp.Draw(x, y);
	return(1);
    }
}

// WINDOW DRAW HANDLER
void NixieDisplay::draw(void)
{
    // DRAW WINDOW
    Fl_Double_Window::draw();

    // DRAW DIGITS ONE AT A TIME
    //    Comp digits that changed.
    //
    int x = XMARGIN;
    for ( int t=0; _dig[t]; t++ )
    {
        if ( _lastdig[t] ) _DrawDigitComp(_dig[t], _lastdig[t], x, YMARGIN);
	else               _DrawDigitComp(_dig[t],     _dig[t], x, YMARGIN);

	x += FONTWIDTH + 2;
	_lastdig[t] = _dig[t];
    }
}

#ifdef STANDALONE
int main()
{
    NixieDisplay nix(500, 45);
    nix.SetString("HELLO WORLD");
    nix.show();
    return(Fl::run());
}
#endif /*STANDALONE*/
